@@ -22,6 +22,7 @@ try:
         valuation,
         get_query_count,
         finance,
+        get_index_stocks,
     )
 
     # 本地导入
@@ -34,7 +35,7 @@ except ImportError:
 from xalpha.info import fundinfo, mfundinfo
 from xalpha.cons import rget, rpost, rget_json, rpost_json, yesterday, opendate
 from xalpha.provider import data_source
-from xalpha.exceptions import DataPossiblyWrong
+from xalpha.exceptions import DataPossiblyWrong, ParserFailure
 
 
 thismodule = sys.modules[__name__]
@@ -51,7 +52,19 @@ def tomorrow_ts():
     return dto.timestamp()
 
 
+def has_weekday(start, end):
+    for d in pd.date_range(start, end):
+        if d.weekday() < 5:
+            return True
+    return False
+
+
 def get_token():
+    """
+    获取雪球的验权 token，匿名也可获取，而且似乎永远恒定
+
+    :return:
+    """
     r = rget("https://xueqiu.com", headers={"user-agent": "Mozilla"})
     return r.cookies["xq_a_token"]
 
@@ -78,7 +91,7 @@ def ts2pdts(ts):
     )  # 雪球美股数据时间戳是美国0点，按北京时区换回时间后，把时分秒扔掉就重合了
 
 
-def get_xueqiu(code, count):
+def get_historical_fromxq(code, count):
     r = get_history(code=code, prefix="", count=count, token=get_token())
     df = pd.DataFrame(data=r["data"]["item"], columns=r["data"]["column"])
     df["date"] = (df["timestamp"]).apply(ts2pdts)  # reset hours to zero
@@ -268,7 +281,7 @@ def get_fund(code):
 
 
 # this is the most elegant approach to dispatch get_daily, the definition can be such simple
-# you actually don't need to bother on start end blah, it is all taken care of by ``cahcedio``
+# you actually don't need to bother on start end blah, everything is taken care of by ``cahcedio``
 @data_source("jq")
 def get_fundshare_byjq(code, **kws):
     code = _inverse_convert_code(code)
@@ -328,7 +341,7 @@ selectedModule=PerformanceGraphView&selectedSubModule=Graph\
 def get_historical_frombb(code, start=None, end=None, **kws):
     """
     https://www.bloomberg.com/ 数据源, 试验性支持。
-    似乎有很严格的 IP 封禁措施, 且最新数据更新滞后，似乎难以支持 T-1 净值预测
+    似乎有很严格的 IP 封禁措施, 且最新数据更新滞后，似乎难以支持 T-1 净值预测。强烈建议从英为或雅虎能找到的标的，不要用彭博源，该 API 只能作为 last resort。
 
     :param code:
     :param start:
@@ -370,10 +383,61 @@ timeframe={years}&period=daily&volumePeriod=daily".format(
     return df
 
 
+def get_historical_fromyh(code, start=None, end=None):
+    """
+    雅虎财经数据源，支持数据丰富，不限于美股。但存在部分历史数据确实 NAN 或者周末进入交易日的现象，可能数据需要进一步清洗和处理。
+
+    :param code:
+    :param start:
+    :param end:
+    :return:
+    """
+    if code.startswith("YH-"):
+        code = code[3:]
+    start_obj = dt.datetime.strptime(start, "%Y%m%d")
+    today_obj = dt.datetime.now()
+    fromnow = (today_obj - start_obj).days
+    if fromnow < 20:
+        range_ = "1mo"
+    elif fromnow < 50:
+        range_ = "3mo"
+    elif fromnow < 150:
+        range_ = "6mo"
+    elif fromnow < 300:
+        range_ = "1y"
+    elif fromnow < 600:
+        range_ = "2y"
+    elif fromnow < 1500:
+        range_ = "5y"
+    else:
+        range_ = "10y"
+    url = "https://query1.finance.yahoo.com/v8\
+/finance/chart/{code}?region=US&lang=en-US&includePrePost=false\
+&interval=1d&range={range_}&corsDomain=finance.yahoo.com&.tsrc=finance".format(
+        code=code, range_=range_
+    )
+    # 该 API 似乎也支持起止时间选择参数，period1=1427500800&period2=1585353600
+    # 也可直接从历史数据页面爬取： https://finance.yahoo.com/quote/CSGOLD.SW/history?period1=1427500800&period2=1585353600&interval=1d&filter=history&frequency=1d
+    r = rget_json(url)
+    data = {}
+    datel = []
+    for t in r["chart"]["result"][0]["timestamp"]:
+        t = dt.datetime.fromtimestamp(t)
+        if t.second != 0:
+            t -= dt.timedelta(hours=8)
+        datel.append(t.replace(tzinfo=None, hour=0, minute=0, second=0, microsecond=0))
+
+    data["date"] = datel
+    for k in ["close", "open", "high", "low"]:
+        data[k] = r["chart"]["result"][0]["indicators"]["quote"][0][k]
+    df = pd.DataFrame(data)
+    return df
+
+
 def _get_daily(code, start=None, end=None, prev=365, _from=None, wrapper=True, **kws):
     """
     universal fetcher for daily historical data of literally everything has a value in market.
-    数据来源包括天天基金，雪球，英为财情，外汇局官网，聚宽，标普官网，bloomberg 等。
+    数据来源包括天天基金，雪球，英为财情，外汇局官网，聚宽，标普官网，bloomberg，雅虎财经等。
 
     :param code: str.
 
@@ -393,7 +457,7 @@ def _get_daily(code, start=None, end=None, prev=365, _from=None, wrapper=True, *
 
             8. 形如 peb-000807.XSHG 或 peb-SH000807 格式的数据，可以返回每周的指数估值情况，需要 enable 聚宽数据源方可查看。
 
-            9. 形如 iw-000807.XSHG 或 peb-SH000807 格式的数据，可以返回每月的指数成分股和实时权重，需要 enable 聚宽数据源方可查看。
+            9. 形如 iw-000807.XSHG 或 iw-SH000807 格式的数据，可以返回每月的指数成分股和实时权重，需要 enable 聚宽数据源方可查看。
 
             10. 形如 fs-SH501018 格式的数据，可以返回指定场内基金每日份额，需要 enable 聚宽数据源方可查看。
 
@@ -401,6 +465,12 @@ def _get_daily(code, start=None, end=None, prev=365, _from=None, wrapper=True, *
 
             12. 形如 BB-FGERBIU:ID 格式的数据，对应网页 https://www.bloomberg.com/quote/FGERBIU:ID，可以返回彭博的数据（最近五年）
 
+            13. 形如 sw-801720 格式的数据，可以返回对应申万行业的历史数据情况，需要 enable 聚宽数据源方可查看。
+
+            14. 形如 teb-SH000300 格式的数据，返回每周指数盈利和净资产总值数据(单位：亿人民币元)，需要 enbale 聚宽数据方可查看。
+
+            15. 形如 YH-CSGOLD.SW 格式的数据，返回雅虎财经标的日线数据（最近十年）。代码来自标的网页 url：https://finance.yahoo.com/quote/CSGOLD.SW。
+
     :param start: str. "20200101", "2020/01/01", "2020-01-01" are all legal. The starting date of daily data.
     :param end: str. format is the same as start. The ending date of daily data.
     :param prev: Optional[int], default 365. If start is not specified, start = end-prev.
@@ -447,13 +517,13 @@ def _get_daily(code, start=None, end=None, prev=365, _from=None, wrapper=True, *
     start_str = start_obj.strftime("%Y/%m/%d")
     end_str = end_obj.strftime("%Y/%m/%d")
 
-    if _from in ["cninvesting", "investing", "default"]:
+    if _from in ["cninvesting", "investing", "default", "IN"]:
         df = get_cninvesting(code, start_str, end_str)
         df = prettify(df)
-    elif _from in ["xueqiu", "xq", "snowball"]:
-        df = get_xueqiu(code, count)
+    elif _from in ["xueqiu", "xq", "snowball", "XQ"]:
+        df = get_historical_fromxq(code, count)
         df = prettify(df)
-    elif _from in ["zhongjianjia", "zjj", "chinamoney"]:
+    elif _from in ["zhongjianjia", "zjj", "chinamoney", "ZJJ"]:
         df = get_rmb(start, end, prev, currency=code)
     elif _from in ["ttjj", "tiantianjijin", "xalpha", "eastmoney"]:
         df = get_fund(code)
@@ -473,6 +543,18 @@ def _get_daily(code, start=None, end=None, prev=365, _from=None, wrapper=True, *
     elif _from == "BB":
         df = get_historical_frombb(code, start=start, end=end)
 
+    elif _from == "sw":
+        df = get_sw_from_jq(code, start=start, end=end)
+
+    elif _from == "teb":
+        df = get_teb_range(code, start=start, end=end)
+
+    elif _from == "YH":
+        df = get_historical_fromyh(code, start=start, end=end)
+
+    else:
+        raise ParserFailure("no such data source: %s" % _from)
+
     if wrapper or len(df) == 0:
         return df
     else:
@@ -656,6 +738,7 @@ def cached(s):
     def cached_start(f):
         @wraps(f)
         def wrapper(*args, **kws):
+            print("cached function is deprecated, please instead use cachedio")
             if args:
                 code = args[0]
             else:
@@ -787,12 +870,14 @@ def cachedio(**ioconf):
                             kws["end"] = (
                                 df0.iloc[0][date] - pd.Timedelta(days=1)
                             ).strftime("%Y%m%d")
-                            df1 = f(*args, **kws)
-                            if len(df1) > 0:
-                                df1 = df1[df1["date"] <= kws["end"]]
-                            if len(df1) > 0:
-                                is_changed = True
-                                df0 = df1.append(df0, ignore_index=True)
+                            if has_weekday(kws["start"], kws["end"]):
+                                # 考虑到海外市场的不同情况，不用 opendate 判断，采取保守型判别
+                                df1 = f(*args, **kws)
+                                if len(df1) > 0:
+                                    df1 = df1[df1["date"] <= kws["end"]]
+                                if len(df1) > 0:
+                                    is_changed = True
+                                    df0 = df1.append(df0, ignore_index=True)
                         # 向后延拓
                         if df0.iloc[-1][date] < end_obj:
                             if len(df0[df0["date"] == df0.iloc[-1]["date"]]) == 1:
@@ -802,14 +887,18 @@ def cachedio(**ioconf):
                                     df0.iloc[-1][date] + dt.timedelta(days=1)
                                 ).strftime("%Y%m%d")
                             kws["end"] = end_str
-                            df2 = f(*args, **kws)
-                            if len(df2) > 0:
-                                df2 = df2[df2["date"] >= kws["start"]]
-                            if len(df2) > 0:
-                                is_changed = True
-                                if len(df0[df0["date"] == df0.iloc[-1]["date"]]) == 1:
-                                    df0 = df0.iloc[:-1]
-                                df0 = df0.append(df2, ignore_index=True)
+                            if has_weekday(kws["start"], kws["end"]):
+                                df2 = f(*args, **kws)
+                                if len(df2) > 0:
+                                    df2 = df2[df2["date"] >= kws["start"]]
+                                if len(df2) > 0:
+                                    is_changed = True
+                                    if (
+                                        len(df0[df0["date"] == df0.iloc[-1]["date"]])
+                                        == 1
+                                    ):
+                                        df0 = df0.iloc[:-1]
+                                    df0 = df0.append(df2, ignore_index=True)
                             # 注意这里抹去更新了原有最后一天的缓存，这是因为日线最新一天可能有实时数据污染
 
                     except (FileNotFoundError, exc.ProgrammingError, KeyError):
@@ -959,6 +1048,50 @@ def get_peb(index, date=None, table=False):
     return {"pe": round(100.0 / tote, 3), "pb": round(100.0 / totb, 3)}
 
 
+@data_source("jq")
+def get_sw_from_jq(code, start=None, end=None, **kws):
+    """
+
+    :param code: str. eg. 801180 申万行业指数
+    :param start:
+    :param end:
+    :param kws:
+    :return:
+    """
+    df = finance.run_query(
+        query(finance.SW1_DAILY_VALUATION)
+        .filter(finance.SW1_DAILY_VALUATION.date >= start)
+        .filter(finance.SW1_DAILY_VALUATION.date <= end)
+        .filter(finance.SW1_DAILY_VALUATION.code == code)
+        .order_by(finance.SW1_DAILY_VALUATION.date.asc())
+    )
+    df["date"] = pd.to_datetime(df["date"])
+    return df
+
+
+def get_teb(code, date):
+    if len(code.split(".")) != 2:
+        code = _inverse_convert_code(code)
+    sl = get_index_stocks(code, date=date)
+    df = get_fundamentals(query(valuation).filter(valuation.code.in_(sl)), date=date)
+    df["e"] = df["market_cap"] / df["pe_ratio"]
+    df["b"] = df["market_cap"] / df["pb_ratio"]
+    return {"e": df["e"].sum(), "b": df["b"].sum()}  # 亿人民币
+
+
+def get_teb_range(code, start, end):
+    if len(code.split(".")) != 2:
+        code = _inverse_convert_code(code)
+    data = {"date": [], "e": [], "b": []}
+    for d in pd.date_range(start, end, freq="W-FRI"):
+        data["date"].append(d)
+        r = get_teb(code, d.strftime("%Y-%m-%d"))
+        data["e"].append(r["e"])
+        data["b"].append(r["b"])
+    df = pd.DataFrame(data)
+    return df
+
+
 def _convert_code(code):
     """
     将聚宽形式的代码转化为 xalpha 形式
